﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;
using System.Threading.Tasks;

namespace TestLibrary
{
    /// <summary>
    /// 비트맵 헬퍼
    /// </summary>
    public static class BitmapHelper
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 비트맵 구하기 - GetBitmap(fftList, colorMap, intensity, db, dbScale, roll, rollOffset)

        /// <summary>
        /// 비트맵 구하기
        /// </summary>
        /// <param name="fftList">FFT 리스트</param>
        /// <param name="colorMap">색상 맵</param>
        /// <param name="intensity">강도</param>
        /// <param name="db">dB 적용 여부</param>
        /// <param name="dbScale">dB 스케일</param>
        /// <param name="roll">롤 여부</param>
        /// <param name="rollOffset">롤 오프셋</param>
        /// <returns>비트맵</returns>
        public static Bitmap GetBitmap
        (
            List<double[]> fftList,
            ColorMap       colorMap,
            double         intensity = 1,
            bool           db = false,
            double         dbScale = 1,
            bool           roll = false,
            int            rollOffset = 0
        )
        {
            if(fftList.Count == 0)
            {
                throw new ArgumentException("This Spectrogram contains no FFTs (likely because no signal was added)");
            }

            int width  = fftList.Count;
            int height = fftList[0].Length;

            Bitmap bitmap = new Bitmap(width, height, PixelFormat.Format8bppIndexed);

            colorMap.SetBitmapPalette(bitmap);

            Rectangle lockRectangle = new Rectangle(0, 0, width, height);

            BitmapData bitmapData = bitmap.LockBits(lockRectangle, ImageLockMode.ReadOnly, bitmap.PixelFormat);

            int stride = bitmapData.Stride;

            byte[] byteArray = new byte[bitmapData.Stride * bitmap.Height];

            Parallel.For
            (
                0,
                width,
                column =>
                {
                    int sourceColumn = column;

                    if(roll)
                    {
                        sourceColumn += width - rollOffset % width;

                        if(sourceColumn >= width)
                        {
                            sourceColumn -= width;
                        }
                    }

                    for(int row = 0; row < height; row++)
                    {
                        double value = fftList[sourceColumn][row];

                        if(db)
                        {
                            value = 20 * Math.Log10(value * dbScale + 1);
                        }

                        value *= intensity;

                        value = Math.Min(value, 255);

                        int bytePosition = (height - 1 - row) * stride + column;

                        byteArray[bytePosition] = (byte)value;
                    }
                }
            );

            Marshal.Copy(byteArray, 0, bitmapData.Scan0, byteArray.Length);

            bitmap.UnlockBits(bitmapData);

            return bitmap;
        }

        #endregion
        #region 수직 스케일 비트맵 구하기 - GetVerticalScaleBitmap(width, setting, hzOffset, tickSize, reduction)

        /// <summary>
        /// 수직 스케일 비트맵 구하기
        /// </summary>
        /// <param name="width">너비</param>
        /// <param name="setting">설정</param>
        /// <param name="hzOffset">HZ 오프셋</param>
        /// <param name="tickSize">틱 크기</param>
        /// <param name="reduction">축소</param>
        /// <returns>수직 스케일 비트맵</returns>
        public static Bitmap GetVerticalScaleBitmap(int width, Setting setting, int hzOffset = 0, int tickSize = 3, int reduction = 1)
        {
            double   hzTick              = 1;
            int      minimumSpacingPixel = 50;
            double[] multiplierArray     = { 2, 2.5, 2 };
            int      multiplier          = 0;

            while(true)
            {
                hzTick *= multiplierArray[multiplier++ % multiplierArray.Length];

                double tickCount         = setting.FrequencySpan / hzTick;
                double pixelBetweenTicks = setting.Height / tickCount;

                if(pixelBetweenTicks >= minimumSpacingPixel * reduction)
                {
                    break;
                }
            }

            Bitmap bitmap = new Bitmap(width, setting.Height / reduction, PixelFormat.Format32bppPArgb);

            using(Graphics     graphics     = Graphics.FromImage(bitmap))
            using(Pen          pen          = new Pen(Color.Black))
            using(SolidBrush   brush        = new SolidBrush(Color.Black))
            using(Font         font         = new Font("나눔고딕코딩", 12))
            using(StringFormat stringFormat = new StringFormat() { LineAlignment = StringAlignment.Center })
            {
                graphics.Clear(Color.White);

                List<double> frequencyList = new List<double>();

                for(double frequency = setting.MinimumFrequency; frequency <= setting.MaximumFrequency; frequency += hzTick)
                {
                    frequencyList.Add(frequency);
                }

                if(frequencyList.Count >= 2)
                {
                    frequencyList.RemoveAt(0);

                    frequencyList.RemoveAt(frequencyList.Count - 1);
                }

                foreach(double frequency in frequencyList)
                {
                    int y = setting.GetPixelY(frequency) / reduction;

                    graphics.DrawLine(pen, 0, y, tickSize, y);

                    graphics.DrawString($"{frequency + hzOffset:N0} Hz", font, brush, tickSize, y, stringFormat);
                }
            }

            return bitmap;
        }

        #endregion
    }
}