﻿using System;
using System.Drawing;
using System.Drawing.Imaging;
using System.Linq;

namespace TestLibrary
{
    /// <summary>
    /// 색상 맵
    /// </summary>
    public class ColorMap
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 아르고 색상 맵 - Argo

        /// <summary>
        /// 아르고 색상 맵
        /// </summary>
        public static ColorMap Argo => new ColorMap(new ArgoColorMap());

        #endregion
        #region 청색 색상 맵 - Blue

        /// <summary>
        /// 청색 색상 맵
        /// </summary>
        public static ColorMap Blue => new ColorMap(new BlueColorMap());

        #endregion
        #region 회색조 색상 맵 - Grayscale

        /// <summary>
        /// 회색조 색상 맵
        /// </summary>
        public static ColorMap Grayscale => new ColorMap(new GrayscaleColorMap());

        #endregion
        #region 회색조 반전 색상 맵 - GrayscaleReversed

        /// <summary>
        /// 회색조 반전 색상 맵
        /// </summary>
        public static ColorMap GrayscaleReversed => new ColorMap(new GrayscaleReversedColorMap());

        #endregion
        #region 녹색 색상 맵 - Green

        /// <summary>
        /// 녹색 색상 맵
        /// </summary>
        public static ColorMap Green => new ColorMap(new GreenColorMap());

        #endregion
        #region 인페르노 색상 맵 - Inferno

        /// <summary>
        /// 인페르노 색상 맵
        /// </summary>
        public static ColorMap Inferno => new ColorMap(new InfernoColorMap());

        #endregion
        #region 로포라 색상 맵 - Lopora

        /// <summary>
        /// 로포라 색상 맵
        /// </summary>
        public static ColorMap Lopora => new ColorMap(new LoporaColorMap());

        #endregion
        #region 마그마 색상 맵 - Magma

        /// <summary>
        /// 마그마 색상 맵
        /// </summary>
        public static ColorMap Magma => new ColorMap(new MagmaColorMap());

        #endregion
        #region 플라스마 색상 맵 - Plasma

        /// <summary>
        /// 플라스마 색상 맵
        /// </summary>
        public static ColorMap Plasma => new ColorMap(new PlasmaColorMap());

        #endregion
        #region 터보 색상 맵 - Turbo

        /// <summary>
        /// 터보 색상 맵
        /// </summary>
        public static ColorMap Turbo => new ColorMap(new TurboColorMap());

        #endregion
        #region 비리디스 색상 맵 - Viridis

        /// <summary>
        /// 비리디스 색상 맵
        /// </summary>
        public static ColorMap Viridis => new ColorMap(new ViridisColorMap());

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Instance
        //////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 색상 맵
        /// </summary>
        private readonly IColorMap colorMap;

        /// <summary>
        /// 명칭
        /// </summary>
        public readonly string Name;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - ColorMap(colorMap)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="colorMap">색상 맵</param>
        public ColorMap(IColorMap colorMap)
        {
            this.colorMap = colorMap ?? new GrayscaleColorMap();

            Name = this.colorMap.GetType().Name;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 색상 맵 배열 구하기 - GetColorMapArray()

        /// <summary>
        /// 색상 맵 배열 구하기
        /// </summary>
        /// <returns>색상 맵 배열</returns>
        public static ColorMap[] GetColorMapArray() => typeof(ColorMap).GetProperties()
                                                                       .Select(x => (ColorMap)x.GetValue(x.Name))
                                                                       .ToArray();

        #endregion
        #region 색상 맵 명칭 배열 구하기 - GetColorMapNameArray()

        /// <summary>
        /// 색상 맵 명칭 배열 구하기
        /// </summary>
        /// <returns>색상 맵 명칭 배열</returns>
        public static string[] GetColorMapNameArray()
        {
            return GetColorMapArray().Select(x => x.Name).ToArray();
        }

        #endregion
        #region 색상 맵 구하기 - GetColorMap(colorMapName)

        /// <summary>
        /// 색상 맵 구하기
        /// </summary>
        /// <param name="colorMapName">색상 맵 명칭</param>
        /// <returns>색상 맵</returns>
        public static ColorMap GetColorMap(string colorMapName)
        {
            foreach(ColorMap colorMap in GetColorMapArray())
            {
                if(string.Equals(colorMap.Name, colorMapName, StringComparison.InvariantCultureIgnoreCase))
                {
                    return colorMap;
                }
            }

            throw new ArgumentException($"ColorMap does not exist : {colorMapName}");
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Instance
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 문자열 구하기 - ToString()

        /// <summary>
        /// 문자열 구하기
        /// </summary>
        /// <returns>문자열</returns>
        public override string ToString()
        {
            return $"ColorMap {Name}";
        }

        #endregion
        #region RGB 구하기 - GetRGB(value)

        /// <summary>
        /// RGB 구하기
        /// </summary>
        /// <param name="value">값</param>
        /// <returns>RGB 튜플</returns>
        public (byte r, byte g, byte b) GetRGB(byte value)
        {
            return this.colorMap.GetRGB(value);
        }

        #endregion
        #region RGB 구하기 - GetRGB(fraction)

        /// <summary>
        /// RGB 구하기
        /// </summary>
        /// <param name="fraction">분수</param>
        /// <returns>RGB 튜플</returns>
        public (byte r, byte g, byte b) GetRGB(double fraction)
        {
            fraction = Math.Max(fraction, 0);
            fraction = Math.Min(fraction, 1);

            return this.colorMap.GetRGB((byte)(fraction * 255));
        }

        #endregion
        #region 32비트 정수 구하기 - GetInt32(value)

        /// <summary>
        /// 32비트 정수 구하기
        /// </summary>
        /// <param name="value">값</param>
        /// <returns>32비트 정수</returns>
        public int GetInt32(byte value)
        {
            var (r, g, b) = GetRGB(value);

            return 255 << 24 | r << 16 | g << 8 | b;
        }

        #endregion
        #region 32비트 정수 구하기 - GetInt32(fraction)

        /// <summary>
        /// 32비트 정수 구하기
        /// </summary>
        /// <param name="fraction">분수</param>
        /// <returns>32비트 정수</returns>
        public int GetInt32(double fraction)
        {
            var (r, g, b) = GetRGB(fraction);

            return 255 << 24 | r << 16 | g << 8 | b;
        }

        #endregion
        #region 색상 구하기 - GetColor(value)

        /// <summary>
        /// 색상 구하기
        /// </summary>
        /// <param name="value">값</param>
        /// <returns>색상</returns>
        public Color GetColor(byte value)
        {
            return Color.FromArgb(GetInt32(value));
        }

        #endregion
        #region 색상 구하기 - GetColor(fraction)

        /// <summary>
        /// 색상 구하기
        /// </summary>
        /// <param name="fraction">분수</param>
        /// <returns>색상</returns>
        public Color GetColor(double fraction)
        {
            return Color.FromArgb(GetInt32(fraction));
        }

        #endregion
        #region 비트맵 팔레트 설정하기 - SetBitmapPalette(bitmap)

        /// <summary>
        /// 비트맵 팔레트 설정하기
        /// </summary>
        /// <param name="bitmap">비트맵</param>
        public void SetBitmapPalette(Bitmap bitmap)
        {
            ColorPalette palette = bitmap.Palette;

            for(int i = 0; i < 256; i++)
            {
                palette.Entries[i] = GetColor((byte)i);
            }

            bitmap.Palette = palette;
        }

        #endregion
    }
}