﻿using System;

using FftSharp;
using FftSharp.Windows;

namespace TestLibrary
{
    /// <summary>
    /// 설정
    /// </summary>
    public class Setting
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region Field

        /// <summary>
        /// 샘플 비율
        /// </summary>
        public readonly int SampleRate;

        /// <summary>
        /// FFT 크기
        /// </summary>
        public readonly int FFTSize;

        /// <summary>
        /// FFT 길이 (단위 : 초)
        /// </summary>
        public readonly double FFTLength;

        /// <summary>
        /// 주파수 나이퀴스트
        /// </summary>
        public readonly double FrequencyNyquist;

        /// <summary>
        /// 픽셀당 HZ
        /// </summary>
        public readonly double HZPerPixel;

        /// <summary>
        /// HZ당 픽셀
        /// </summary>
        public readonly double PixelPerHZ;

        /// <summary>
        /// FFT 인덱스 1
        /// </summary>
        public readonly int FFTIndex1;

        /// <summary>
        /// FFT 인덱스 2
        /// </summary>
        public readonly int FFTIndex2;

        /// <summary>
        /// 최소 주파수
        /// </summary>
        public readonly double MinimumFrequency;

        /// <summary>
        /// 최대 주파수
        /// </summary>
        public readonly double MaximumFrequency;

        /// <summary>
        /// 주파수 범위
        /// </summary>
        public readonly double FrequencySpan;

        /// <summary>
        /// 높이
        /// </summary>
        public readonly int Height;

        /// <summary>
        /// HZ 오프셋
        /// </summary>
        public int HZOffset;

        /// <summary>
        /// 윈도우 배열
        /// </summary>
        public readonly double[] WindowArray;

        /// <summary>
        /// 단계 크기
        /// </summary>
        public readonly int StepSize;

        /// <summary>
        /// 단계 길이 (단위 : 초)
        /// </summary>
        public readonly double StepLength;

        /// <summary>
        /// 단계 오버랩 (분수)
        /// </summary>
        public readonly double StepOverlapFraction;

        /// <summary>
        /// 단계 오버랩 (단위 : 초)
        /// </summary>
        public readonly double StepOverlap;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - Setting(sampleRate, fftSize, stepSize, minimumFrequency, maximumFrequency, hzOffset)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="sampleRate">샘플 비율</param>
        /// <param name="fftSize">FFT 크기</param>
        /// <param name="stepSize">단계 크기</param>
        /// <param name="minimumFrequency">최소 주파수</param>
        /// <param name="maximumFrequency">최대 주파수</param>
        /// <param name="hzOffset">HZ 오프셋</param>
        public Setting(int sampleRate, int fftSize, int stepSize, double minimumFrequency, double maximumFrequency, int hzOffset)
        {
            if(Transform.IsPowerOfTwo(fftSize) == false)
            {
                throw new ArgumentException("FFT size must be a power of 2");
            }

            SampleRate          = sampleRate;
            FFTSize             = fftSize;
            StepSize            = stepSize;
            FFTLength           = (double)fftSize / sampleRate;
            minimumFrequency    = Math.Max(minimumFrequency, 0);
            FrequencyNyquist    = sampleRate / 2;
            HZPerPixel          = (double)sampleRate / fftSize;
            PixelPerHZ          = (double)fftSize / sampleRate;
            FFTIndex1           = (minimumFrequency == 0) ? 0 : (int)(minimumFrequency / HZPerPixel);
            FFTIndex2           = (maximumFrequency >= FrequencyNyquist) ? fftSize / 2 : (int)(maximumFrequency / HZPerPixel);
            Height              = FFTIndex2 - FFTIndex1;
            MinimumFrequency    = FFTIndex1 * HZPerPixel;
            MaximumFrequency    = FFTIndex2 * HZPerPixel;
            FrequencySpan       = MaximumFrequency - MinimumFrequency;
            HZOffset            = hzOffset;
            StepLength          = (double)StepSize / sampleRate;
            WindowArray         = new Hanning().Create(fftSize);
            StepOverlap         = FFTLength - StepLength;
            StepOverlapFraction = StepOverlap / FFTLength;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 픽셀 Y 구하기 - GetPixelY(frequency)

        /// <summary>
        /// 픽셀 Y 구하기
        /// </summary>
        /// <param name="frequency">주파수</param>
        /// <returns>픽셀 Y</returns>
        public int GetPixelY(double frequency)
        {
            return (int)(Height - (frequency - MinimumFrequency + HZPerPixel) * PixelPerHZ - 1);
        }

        #endregion
    }
}