﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Text;

using FftSharp;

namespace TestLibrary
{
    /// <summary>
    /// 스펙트로그램 파일 포맷
    /// </summary>
    public class SpectrogramFileFormat
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region Field

        /// <summary>
        /// 메이저 버전
        /// </summary>
        public readonly byte MajorVersion = 1;

        /// <summary>
        /// 마이너 버전
        /// </summary>
        public readonly byte MinorVersion = 1;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 파일 경로 - FilePath

        /// <summary>
        /// 파일 경로
        /// </summary>
        public string FilePath { get; private set; }

        #endregion
        #region 샘플 비율 - SampleRate

        /// <summary>
        /// 샘플 비율
        /// </summary>
        public int SampleRate { get; private set; }

        #endregion
        #region 단계 크기 - StepSize

        /// <summary>
        /// 단계 크기
        /// </summary>
        public int StepSize { get; private set; }

        #endregion
        #region 너비 - Width

        /// <summary>
        /// 너비
        /// </summary>
        public int Width { get; private set; }

        #endregion
        #region FFT 크기 - FFTSize

        /// <summary>
        /// FFT 크기
        /// </summary>
        public int FFTSize { get; private set; }

        #endregion
        #region FFT 첫번째 인덱스 - FFTFirstIndex

        /// <summary>
        /// FFT 첫번째 인덱스
        /// </summary>
        public int FFTFirstIndex { get; private set; }

        #endregion
        #region 높이 - Height

        /// <summary>
        /// 높이
        /// </summary>
        public int Height { get; private set; }

        #endregion
        #region HZ 오프셋 - HZOffset

        /// <summary>
        /// HZ 오프셋
        /// </summary>
        public int HZOffset { get; private set; }

        #endregion
        #region MEL BIN 카운트 - MELBINCount

        /// <summary>
        /// MEL BIN 카운트
        /// </summary>
        public int MELBINCount { get; private set; }

        #endregion
        #region 데시빌 카운트 - DecibelCount

        /// <summary>
        /// 데시빌 카운트
        /// </summary>
        public bool DecibelCount { get; private set; }

        #endregion
        #region MEL 여부 - IsMEL

        /// <summary>
        /// MEL 여부
        /// </summary>
        public bool IsMEL
        {
            get
            {
                return MELBINCount > 0;
            }
        }

        #endregion
        #region FFT 리스트 - FFTList

        /// <summary>
        /// FFT 리스트
        /// </summary>
        public List<double[]> FFTList { get; private set; }

        #endregion
        #region 이미지 너비 - ImageWidth

        /// <summary>
        /// 이미지 너비
        /// </summary>
        public int ImageWidth { get { return (FFTList is null) ? 0 : FFTList.Count; } }

        #endregion
        #region 이미지 높이 - ImageHeight

        /// <summary>
        /// 이미지 높이
        /// </summary>
        public int ImageHeight { get { return (FFTList is null) ? 0 : FFTList[0].Length; } }

        #endregion
        #region 시간 배열 - TimeArray

        /// <summary>
        /// 시간 배열
        /// </summary>
        public double[] TimeArray { get; private set; }

        #endregion
        #region 주파수 배열 - FrequencyArray

        /// <summary>
        /// 주파수 배열
        /// </summary>
        public double[] FrequencyArray { get; private set; }

        #endregion
        #region MEL 배열 - MELArray

        /// <summary>
        /// MEL 배열
        /// </summary>
        public double[] MELArray { get; private set; }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - SpectrogramFileFormat()

        /// <summary>
        /// 생성자
        /// </summary>
        public SpectrogramFileFormat()
        {
        }

        #endregion
        #region 생성자 - SpectrogramFileFormat(filePath)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="filePath">파일 경로</param>
        public SpectrogramFileFormat(string filePath)
        {
            Load(filePath);

            CalculateTimes();

            CalculateFrequencies();
        }

        #endregion
        #region 생성자 - SpectrogramFileFormat(generator, melBinCount)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="generator">스펙트로그램 제너레이터</param>
        /// <param name="melBinCount">MEL BIN 카운트</param>
        public SpectrogramFileFormat(SpectrogramGenerator generator, int melBinCount = 0)
        {
            SampleRate    = generator.SampleRate;
            StepSize      = generator.StepSize;
            Width         = generator.Width;
            FFTSize       = generator.FFTSize;
            FFTFirstIndex = generator.NextColumnIndex;
            Height        = generator.Height;
            HZOffset      = generator.HZOffset;
            MELBINCount   = melBinCount;
            FFTList       = (melBinCount > 0) ? generator.GetMELFFTArrayList(melBinCount) : generator.GetFFTArrayList();

            CalculateTimes();

            CalculateFrequencies();
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 문자열 구하기 - ToString()

        /// <summary>
        /// 문자열 구하기
        /// </summary>
        /// <returns>문자열</returns>
        public override string ToString()
        {
            return $"SFF {ImageWidth}X{ImageHeight}";
        }

        #endregion
        #region 비트맵 구하기 - GetBitmap(colorMap, intensity, db)

        /// <summary>
        /// 비트맵 구하기
        /// </summary>
        /// <param name="colorMap">색상 맵</param>
        /// <param name="intensity">강도</param>
        /// <param name="db">데시빌 여부</param>
        /// <returns>비트맵</returns>
        public Bitmap GetBitmap(ColorMap colorMap = null, double intensity = 1, bool db = false)
        {
            colorMap = colorMap ?? ColorMap.Viridis;

            return BitmapHelper.GetBitmap(FFTList, colorMap, intensity, db);
        }

        #endregion
        #region 로드하기 - Load(filePath)

        /// <summary>
        /// 로드하기
        /// </summary>
        /// <param name="filePath">파일 경로</param>
        public void Load(string filePath)
        {
            FilePath = Path.GetFullPath(filePath);

            byte[] byteArray = File.ReadAllBytes(filePath);

            int magicNumber = BitConverter.ToInt32(byteArray, 0);

            if(magicNumber != 1179014099)
            {
                throw new InvalidDataException("not a valid SFF file");
            }

            SampleRate    = BitConverter.ToInt32(byteArray, 42);
            StepSize      = BitConverter.ToInt32(byteArray, 46);
            Width         = BitConverter.ToInt32(byteArray, 50);
            FFTSize       = BitConverter.ToInt32(byteArray, 54);
            FFTFirstIndex = BitConverter.ToInt32(byteArray, 58);
            Height        = BitConverter.ToInt32(byteArray, 62);
            HZOffset      = BitConverter.ToInt32(byteArray, 66);
            MELBINCount   = BitConverter.ToInt32(byteArray, 84);

            byte valueCountPerPoint = byteArray[70];
            bool isComplex          = valueCountPerPoint == 2;

            if(isComplex)
            {
                throw new NotImplementedException("complex data is not yet supported");
            }

            byte byteCountPerValue = byteArray[71];

            DecibelCount = byteArray[72] == 1;

            int firstDataByte = (int)BitConverter.ToUInt32(byteArray, 80);

            MELBINCount = BitConverter.ToInt32(byteArray, 84);

            int FFTHeight = BitConverter.ToInt32(byteArray, 88);
            int FFTWidth  = BitConverter.ToInt32(byteArray, 92);

            FFTList = new List<double[]>();

            int byteCountPerPoint  = byteCountPerValue * valueCountPerPoint;
            int byteCountPerColumn = FFTHeight * byteCountPerPoint;

            for(int x = 0; x < FFTWidth; x++)
            {
                FFTList.Add(new double[FFTHeight]);

                int columnOffset = byteCountPerColumn * x;

                for(int y = 0; y < FFTHeight; y++)
                {
                    int rowOffset   = y * byteCountPerPoint;
                    int valueOffset = firstDataByte + columnOffset + rowOffset;

                    double value = BitConverter.ToDouble(byteArray, valueOffset);

                    FFTList[x][y] = value;
                }
            }
        }

        #endregion
        #region 저장하기 - Save(filePath)

        /// <summary>
        /// 저장하기
        /// </summary>
        /// <param name="filePath">파일 경로</param>
        public void Save(string filePath)
        {
            FilePath = Path.GetFullPath(filePath);

            byte[] headerArray = new byte[256];

            headerArray[0] = 211;
            headerArray[1] = (byte)'S';
            headerArray[2] = (byte)'F';
            headerArray[3] = (byte)'F';
            headerArray[4] = (byte)'\r';
            headerArray[5] = (byte)'\n';
            headerArray[6] = (byte)' ';
            headerArray[7] = (byte)'\n';

            int magicNumber = BitConverter.ToInt32(headerArray, 0);

            if(magicNumber != 1179014099)
            {
                throw new InvalidDataException("magic number for SFF files is 1179014099");
            }

            string fileInfo = $"SFF {MajorVersion}.{MinorVersion}\r\n";

            byte[] fileInfoByteArray = Encoding.UTF8.GetBytes(fileInfo);

            if(fileInfoByteArray.Length > 32)
            {
                throw new InvalidDataException("file info cannot exceed 32 bytes");
            }

            Array.Copy(fileInfoByteArray, 0, headerArray, 8, fileInfoByteArray.Length);

            headerArray[40] = MajorVersion;
            headerArray[41] = MinorVersion;

            Array.Copy(BitConverter.GetBytes(SampleRate), 0, headerArray, 42, 4);
            Array.Copy(BitConverter.GetBytes(StepSize  ), 0, headerArray, 46, 4);
            Array.Copy(BitConverter.GetBytes(Width     ), 0, headerArray, 50, 4);

            Array.Copy(BitConverter.GetBytes(FFTSize      ), 0, headerArray, 54, 4);
            Array.Copy(BitConverter.GetBytes(FFTFirstIndex), 0, headerArray, 58, 4);
            Array.Copy(BitConverter.GetBytes(Height       ), 0, headerArray, 62, 4);
            Array.Copy(BitConverter.GetBytes(HZOffset     ), 0, headerArray, 66, 4);

            byte valueCountPerPoint = 1;
            byte byteCountPerValue  = 8;
            byte decibelUnitCount   = 0;
            byte dataExtraByte      = 0;

            headerArray[70] = valueCountPerPoint;
            headerArray[71] = byteCountPerValue;
            headerArray[72] = decibelUnitCount;
            headerArray[73] = dataExtraByte;

            Array.Copy(BitConverter.GetBytes(MELBINCount), 0, headerArray, 84, 4);
            Array.Copy(BitConverter.GetBytes(ImageHeight), 0, headerArray, 88, 4);
            Array.Copy(BitConverter.GetBytes(ImageWidth ), 0, headerArray, 92, 4);

            int firstDataByte = headerArray.Length;

            Array.Copy(BitConverter.GetBytes(firstDataByte), 0, headerArray, 80, 4);

            int dataPointCount    = ImageHeight * ImageWidth;
            int byteCountPerPoint = byteCountPerValue * valueCountPerPoint;

            byte[] fileByteArray = new byte[headerArray.Length + dataPointCount * byteCountPerPoint];

            Array.Copy(headerArray, 0, fileByteArray, 0, headerArray.Length);

            int byteCountPerColumn = ImageHeight * byteCountPerPoint;

            for(int x = 0; x < ImageWidth; x++)
            {
                int columnOffset = byteCountPerColumn * x;

                for(int y = 0; y < ImageHeight; y++)
                {
                    int rowOffset   = y * byteCountPerPoint;
                    int valueOffset = firstDataByte + columnOffset + rowOffset;

                    double value = FFTList[x][y];

                    Array.Copy(BitConverter.GetBytes(value), 0, fileByteArray, valueOffset, 8);
                }
            }

            File.WriteAllBytes(filePath, fileByteArray);
        }

        #endregion
        #region 픽셀 정보 구하기 - GetPixelInformation(x, y)

        /// <summary>
        /// 픽셀 정보 구하기
        /// </summary>
        /// <param name="x">X</param>
        /// <param name="y">Y</param>
        /// <returns>(시간(단위 : 초), 주파수(단위 : HZ), 진도) 튜플</returns>
        public (double timeSecond, double frequencyHZ, double magnitude) GetPixelInformation(int x, int y)
        {
            double timeSecond = (double)x * StepSize / SampleRate;

            double maximumFrequency = SampleRate / 2;
            double maximumMEL       = Transform.MelFromFreq(maximumFrequency);
            double fraction         = (ImageHeight - y) / (double)ImageHeight;
            double frequency        = IsMEL ? Transform.MelToFreq(fraction * maximumMEL) : fraction * maximumFrequency;

            double magnitude = double.NaN;

            try
            {
                magnitude = FFTList[x][ImageHeight - y - 1];
            }
            catch
            {
            }

            return (timeSecond, frequency, magnitude);
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 시간 계산하기 - CalculateTimes()

        /// <summary>
        /// 시간 계산하기
        /// </summary>
        private void CalculateTimes()
        {
            TimeArray = new double[ImageWidth];

            double step = (double)StepSize / SampleRate;

            for(int i = 0; i < ImageWidth; i++)
            {
                TimeArray[i] = i * step;
            }
        }

        #endregion
        #region 주파수 계산하기 - CalculateFrequencies()

        /// <summary>
        /// 주파수 계산하기
        /// </summary>
        private void CalculateFrequencies()
        {
            FrequencyArray = new double[ImageHeight];
            MELArray       = new double[ImageHeight];

            double maximumFrequency = SampleRate / 2;
            double maximumMEL       = Transform.MelFromFreq(maximumFrequency);

            for(int y = 0; y < ImageHeight; y++)
            {
                double fraction = (ImageHeight - y) / (double)ImageHeight;

                if(IsMEL)
                {
                    MELArray[y]       = fraction * maximumMEL;
                    FrequencyArray[y] = Transform.MelToFreq(MELArray[y]);
                }
                else
                {
                    FrequencyArray[y] = fraction * maximumFrequency;
                    MELArray[y]       = Transform.MelFromFreq(FrequencyArray[y]);
                }
            }
        }

        #endregion
    }
}