﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Threading.Tasks;

using FftSharp;

namespace TestLibrary
{
    /// <summary>
    /// 스펙트로그램 제너레이터
    /// </summary>
    public class SpectrogramGenerator
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region Field

        /// <summary>
        /// 색상 맵
        /// </summary>
        public ColorMap Colormap = ColorMap.Viridis;

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 설정
        /// </summary>
        private readonly Setting Setting;

        /// <summary>
        /// FFT 배열 리스트
        /// </summary>
        private readonly List<double[]> FFTArrayList = new List<double[]>();

        /// <summary>
        /// 미처리 FFT 리스트
        /// </summary>
        private readonly List<double> UnprocessedFFTList;

        /// <summary>
        /// 롤 오프셋
        /// </summary>
        private int rollOffset = 0;

        /// <summary>
        /// 고정 너비
        /// </summary>
        private int fixedWidth = 0;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 너비 - Width

        /// <summary>
        /// 너비
        /// </summary>
        public int Width { get => FFTArrayList.Count; }

        #endregion
        #region 높이 - Height

        /// <summary>
        /// 높이
        /// </summary>
        public int Height { get => Setting.Height; }

        #endregion
        #region FFT 크기 - FFTSize

        /// <summary>
        /// FFT 크기
        /// </summary>
        public int FFTSize { get => Setting.FFTSize; }

        #endregion
        #region 픽셀당 HZ - HZPerPixel

        /// <summary>
        /// 픽셀당 HZ
        /// </summary>
        public double HZPerPixel { get => Setting.HZPerPixel; }

        #endregion
        #region 픽셀당 초 카운트 - SecondCountPerPixel

        /// <summary>
        /// 픽셀당 초 카운트
        /// </summary>
        public double SecondCountPerPixel { get => Setting.StepLength; }

        #endregion
        #region 처리할 FFT 카운트 - FFTCountToProcess

        /// <summary>
        /// 처리할 FFT 카운트
        /// </summary>
        public int FFTCountToProcess { get => (UnprocessedFFTList.Count - Setting.FFTSize) / Setting.StepSize; }

        #endregion
        #region 처리 완료 FFT 카운트 - FFTCountProcessed

        /// <summary>
        /// 처리 완료 FFT 카운트
        /// </summary>
        public int FFTCountProcessed { get; private set; }

        #endregion
        #region 다음 컬럼 인덱스 - NextColumnIndex

        /// <summary>
        /// 다음 컬럼 인덱스
        /// </summary>
        public int NextColumnIndex { get => (FFTCountProcessed + rollOffset) % Width; }

        #endregion
        #region HZ 오프셋 - HZOffset

        /// <summary>
        /// HZ 오프셋
        /// </summary>
        public int HZOffset
        {
            get => Setting.HZOffset;
            set
            {
                Setting.HZOffset = value;
            }
        }

        #endregion
        #region 샘플 비율 - SampleRate

        /// <summary>
        /// 샘플 비율
        /// </summary>
        public int SampleRate { get => Setting.SampleRate; }

        #endregion
        #region 단계 크기 - StepSize 

        /// <summary>
        /// 단계 크기
        /// </summary>
        public int StepSize { get => Setting.StepSize; }

        #endregion
        #region 최대 주파수 - MaximumFrequency

        /// <summary>
        /// 최대 주파수
        /// </summary>
        public double MaximumFrequency { get => Setting.MaximumFrequency; }

        #endregion
        #region 최소 주파수 - MinimumFrequency

        /// <summary>
        /// 최소 주파수
        /// </summary>
        public double MinimumFrequency { get => Setting.MinimumFrequency; }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - SpectrogramGenerator(sampleRate, fftSize, stepSize, minimumFrequency, maximumFrequency, fixedWidth, hzOffset, initialAudioList)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="sampleRate">샘플 비율</param>
        /// <param name="fftSize">FFT 크기</param>
        /// <param name="stepSize">단계 크기</param>
        /// <param name="minimumFrequency">최소 주파수</param>
        /// <param name="maximumFrequency">최대 주파수</param>
        /// <param name="fixedWidth">고정 너비</param>
        /// <param name="hzOffset">HZ 오프셋</param>
        /// <param name="initialAudioList">초기 오디오 리스트</param>
        public SpectrogramGenerator
        (
            int          sampleRate,
            int          fftSize,
            int          stepSize,
            double       minimumFrequency = 0,
            double       maximumFrequency = double.PositiveInfinity,
            int?         fixedWidth       = null,
            int          hzOffset         = 0,
            List<double> initialAudioList = null
        )
        {
            Setting = new Setting(sampleRate, fftSize, stepSize, minimumFrequency, maximumFrequency, hzOffset);

            UnprocessedFFTList = initialAudioList ?? new List<double>();

            if(fixedWidth.HasValue)
            {
                SetFixedWidth(fixedWidth.Value);
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 문자열 구하기 - ToString()

        /// <summary>
        /// 문자열 구하기
        /// </summary>
        /// <returns>문자열</returns>
        public override string ToString()
        {
            double processedSampleCount = FFTArrayList.Count * Setting.StepSize + Setting.FFTSize;
            double processedSecondCount = processedSampleCount / Setting.SampleRate;
            string processedTime        = (processedSecondCount < 60) ? $"{processedSecondCount:N2} 초" : $"{processedSecondCount / 60.0:N2} 분";

            return $"스펙트로그램 ({Width}, {Height})" +
                   $"\n  수직 ({Height} 픽셀) : {Setting.MinimumFrequency:N0} - {Setting.MaximumFrequency:N0} HZ, " +
                   $"FFT 크기 : {Setting.FFTSize:N0} 샘플 수, {Setting.HZPerPixel:N2} HZ/픽셀" +
                   $"\n  수평 ({Width} 픽셀) : {processedTime}, 윈도우 : {Setting.FFTLength:N2} 초, " +
                   $"단계 : {Setting.StepLength:N2} 초, 오버랩 : {Setting.StepOverlapFraction * 100:N0}%";
        }

        #endregion
        #region 윈도우 설정하기 - SetWindow(newWindowArray)

        /// <summary>
        /// 윈도우 설정하기
        /// </summary>
        /// <param name="newWindowArray">신규 윈도우 배열</param>
        public void SetWindow(double[] newWindowArray)
        {
            if(newWindowArray.Length > Setting.FFTSize)
            {
                throw new ArgumentException("window length cannot exceed FFT size");
            }

            for(int i = 0; i < Setting.FFTSize; i++)
            {
                Setting.WindowArray[i] = 0;
            }

            int offset = (Setting.FFTSize - newWindowArray.Length) / 2;

            Array.Copy(newWindowArray, 0, Setting.WindowArray, offset, newWindowArray.Length);
        }

        #endregion
        #region 추가하기 - Add(audioEnumerable, process)

        /// <summary>
        /// 추가하기
        /// </summary>
        /// <param name="audioEnumerable">오디오 열거 가능형</param>
        /// <param name="process">처리 여부</param>
        public void Add(IEnumerable<double> audioEnumerable, bool process = true)
        {
            UnprocessedFFTList.AddRange(audioEnumerable);

            if(process)
            {
                Process();
            }
        }

        #endregion
        #region 롤 리셋하기 - ResetRoll(offset)

        /// <summary>
        /// 롤 리셋하기
        /// </summary>
        /// <param name="offset">오프셋</param>
        public void ResetRoll(int offset = 0)
        {
            this.rollOffset = -FFTCountProcessed + offset;
        }

        #endregion
        #region 처리하기 - Process()

        /// <summary>
        /// 처리하기
        /// </summary>
        /// <returns>처리 결과</returns>
        public double[][] Process()
        {
            if(FFTCountToProcess < 1)
            {
                return null;
            }

            int newFFTCount = FFTCountToProcess;

            double[][] newFFTArray = new double[newFFTCount][];

            Parallel.For
            (
                0,
                newFFTCount,
                newFFTIndex =>
                {
                    Complex[] bufferArray = new Complex[Setting.FFTSize];

                    int sourceIndex = newFFTIndex * Setting.StepSize;

                    for(int i = 0; i < Setting.FFTSize; i++)
                    {
                        bufferArray[i].Real = UnprocessedFFTList[sourceIndex + i] * Setting.WindowArray[i];
                    }

                    Transform.FFT(bufferArray);

                    newFFTArray[newFFTIndex] = new double[Setting.Height];

                    for(int i = 0; i < Setting.Height; i++)
                    {
                        newFFTArray[newFFTIndex][i] = bufferArray[Setting.FFTIndex1 + i].Magnitude / Setting.FFTSize;
                    }
                }
            );

            foreach(var newFFT in newFFTArray)
            {
                FFTArrayList.Add(newFFT);
            }

            FFTCountProcessed += newFFTArray.Length;

            UnprocessedFFTList.RemoveRange(0, newFFTCount * Setting.StepSize);

            PadOrTrimForFixedWidth();

            return newFFTArray;
        }

        #endregion
        #region MEL FFT 배열 리스트 구하기 - GetMELFFTArrayList(melBINCount)

        /// <summary>
        /// MEL FFT 배열 리스트 구하기
        /// </summary>
        /// <param name="melBINCount">MEL BIN 카운트</param>
        /// <returns>MEL FFT 배열 리스트</returns>
        public List<double[]> GetMELFFTArrayList(int melBINCount)
        {
            if(Setting.MinimumFrequency != 0)
            {
                throw new InvalidOperationException("cannot get Mel spectrogram unless minimum frequency is 0Hz");
            }

            List<double[]> melFFTArrayList = new List<double[]>();

            foreach(double[] fftArray in FFTArrayList)
            {
                melFFTArrayList.Add(Transform.MelScale(fftArray, SampleRate, melBINCount));
            }

            return melFFTArrayList;
        }

        #endregion
        #region 비트맵 구하기 - GetBitmap(intensity, db, dbScale, roll)

        /// <summary>
        /// 비트맵 구하기
        /// </summary>
        /// <param name="intensity">강도</param>
        /// <param name="db">데시빌 여부</param>
        /// <param name="dbScale">데시빌 스케일</param>
        /// <param name="roll">롤 여부</param>
        /// <returns>비트맵</returns>
        public Bitmap GetBitmap(double intensity = 1, bool db = false, double dbScale = 1, bool roll = false) =>
            BitmapHelper.GetBitmap(FFTArrayList, Colormap, intensity, db, dbScale, roll, NextColumnIndex);

        #endregion
        #region MEL 비트맵 구하기 - GetMELBitmap(melBINCount, intensity, db, dbScale, roll)

        /// <summary>
        /// MEL 비트맵 구하기
        /// </summary>
        /// <param name="melBINCount">MEL BIN 카운트</param>
        /// <param name="intensity">강도</param>
        /// <param name="db">데시빌 여부</param>
        /// <param name="dbScale">데시빌 스케일</param>
        /// <param name="roll">롤 여부</param>
        /// <returns>MEL 비트맵</returns>
        public Bitmap GetMELBitmap(int melBINCount = 25, double intensity = 1, bool db = false, double dbScale = 1, bool roll = false) =>
            BitmapHelper.GetBitmap(GetMELFFTArrayList(melBINCount), Colormap, intensity, db, dbScale, roll, NextColumnIndex);

        #endregion
        #region 이미지 저장하기 - SaveImage(filePath, intensity, db, dbScale, roll)

        /// <summary>
        /// 이미지 저장하기
        /// </summary>
        /// <param name="filePath">파일 경로</param>
        /// <param name="intensity">강도</param>
        /// <param name="db">데시빌 여부</param>
        /// <param name="dbScale">데시빌 스케일</param>
        /// <param name="roll">롤 여부</param>
        public void SaveImage(string filePath, double intensity = 1, bool db = false, double dbScale = 1, bool roll = false)
        {
            if(FFTArrayList.Count == 0)
            {
                throw new InvalidOperationException("Spectrogram contains no data. Use Add() to add signal data.");
            }

            string extension = Path.GetExtension(filePath).ToLower();

            ImageFormat imageFormat;

            if(extension == ".bmp")
            {
                imageFormat = ImageFormat.Bmp;
            }
            else if(extension == ".png")
            {
                imageFormat = ImageFormat.Png;
            }
            else if(extension == ".jpg" || extension == ".jpeg")
            {
                imageFormat = ImageFormat.Jpeg;
            }
            else if(extension == ".gif")
            {
                imageFormat = ImageFormat.Gif;
            }
            else
            {
                throw new ArgumentException("unknown file extension");
            }

            BitmapHelper.GetBitmap(FFTArrayList, Colormap, intensity, db, dbScale, roll, NextColumnIndex).Save(filePath, imageFormat);
        }

        #endregion
        #region 최대 비트맵 구하기 - GetMaximumBitmap(intensity, db, dbScale, roll, reduction)

        /// <summary>
        /// 최대 비트맵 구하기
        /// </summary>
        /// <param name="intensity">강도</param>
        /// <param name="db">데시빌 여부</param>
        /// <param name="dbScale">데시빌 스케일</param>
        /// <param name="roll">롤 여부</param>
        /// <param name="reduction">축소</param>
        /// <returns>최대 비트맵</returns>
        public Bitmap GetMaximumBitmap(double intensity = 1, bool db = false, double dbScale = 1, bool roll = false, int reduction = 4)
        {
            List<double[]> fftArrayList = new List<double[]>();

            for(int i = 0; i < FFTArrayList.Count; i++)
            {
                double[] valueArray1 = FFTArrayList[i];
                double[] valueArray2 = new double[valueArray1.Length / reduction];

                for(int j = 0; j < valueArray2.Length; j++)
                {
                    for(int k = 0; k < reduction; k++)
                    {
                        valueArray2[j] = Math.Max(valueArray2[j], valueArray1[j * reduction + k]);
                    }
                }

                fftArrayList.Add(valueArray2);
            }

            return BitmapHelper.GetBitmap(fftArrayList, Colormap, intensity, db, dbScale, roll, NextColumnIndex);
        }

        #endregion
        #region 데이터 저장하기 - SaveData(filePath, melBINCount)

        /// <summary>
        /// 데이터 저장하기
        /// </summary>
        /// <param name="filePath">파일 경로</param>
        /// <param name="melBINCount">MEL BIN 카운트</param>
        public void SaveData(string filePath, int melBINCount = 0)
        {
            if(!filePath.EndsWith(".sff", StringComparison.OrdinalIgnoreCase))
            {
                filePath += ".sff";
            }

            new SpectrogramFileFormat(this, melBINCount).Save(filePath);
        }

        #endregion
        #region 고정 너비 설정하기 - SetFixedWidth(width)

        /// <summary>
        /// 고정 너비 설정하기
        /// </summary>
        /// <param name="width">너비</param>
        public void SetFixedWidth(int width)
        {
            this.fixedWidth = width;

            PadOrTrimForFixedWidth();
        }

        #endregion
        #region 수직 스케일 비트맵 구하기 - GetVerticalScaleBitmap(width, hzOffset, tickSize, reduction)

        /// <summary>
        /// 수직 스케일 비트맵 구하기
        /// </summary>
        /// <param name="width">너비</param>
        /// <param name="hzOffset">HZ 오프셋</param>
        /// <param name="tickSize">틱 크기</param>
        /// <param name="reduction">축소</param>
        /// <returns>수직 스케일 비트맵</returns>
        public Bitmap GetVerticalScaleBitmap(int width, int hzOffset = 0, int tickSize = 3, int reduction = 1)
        {
            return BitmapHelper.GetVerticalScaleBitmap(width, Setting, hzOffset, tickSize, reduction);
        }

        #endregion
        #region 픽셀 Y 구하기 - GetPixelY(frequency, reduction)

        /// <summary>
        /// 픽셀 Y 구하기
        /// </summary>
        /// <param name="frequency">주파수</param>
        /// <param name="reduction">축소</param>
        /// <returns>픽셀 Y</returns>
        public int GetPixelY(double frequency, int reduction = 1)
        {
            int pixelCountFromZeroHZ           = (int)(Setting.PixelPerHZ * frequency / reduction);
            int pixelCountFromMinimumFrequency = pixelCountFromZeroHZ - Setting.FFTIndex1 / reduction + 1;
            int pixelRow                       = Setting.Height / reduction - 1 - pixelCountFromMinimumFrequency;

            return pixelRow - 1;
        }

        #endregion
        #region FFT 배열 리스트 구하기 - GetFFTArrayList()

        /// <summary>
        /// FFT 배열 리스트 구하기
        /// </summary>
        /// <returns>FFT 배열 리스트</returns>
        public List<double[]> GetFFTArrayList()
        {
            return FFTArrayList;
        }

        #endregion
        #region 피크 구하기 - GetPeak(latestFFT)

        /// <summary>
        /// 피크 구하기
        /// </summary>
        /// <param name="latestFFT">최신 FFT</param>
        /// <returns>(주파수 (단위 : HZ), 진도) 튜플</returns>
        public (double frequencyHZ, double magnitude) GetPeak(bool latestFFT = true)
        {
            if(FFTArrayList.Count == 0)
            {
                return (double.NaN, double.NaN);
            }

            if(latestFFT == false)
            {
                throw new NotImplementedException("peak of mean of all FFTs not yet supported");
            }

            double[] frequencyList = FFTArrayList[FFTArrayList.Count - 1];

            int peakIndex = 0;

            double peakMagnitude = 0;

            for(int i = 0; i < frequencyList.Length; i++)
            {
                if(frequencyList[i] > peakMagnitude)
                {
                    peakMagnitude = frequencyList[i];

                    peakIndex = i;
                }
            }

            double maximumFrequency      = SampleRate / 2;
            double peakFrequencyFraction = peakIndex / (double)frequencyList.Length;
            double peakFrequencyHZ       = maximumFrequency * peakFrequencyFraction;

            return (peakFrequencyHZ, peakMagnitude);
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 고정 너비용 제거 또는 채우기 - PadOrTrimForFixedWidth()

        /// <summary>
        /// 고정 너비용 제거 또는 채우기
        /// </summary>
        private void PadOrTrimForFixedWidth()
        {
            if(fixedWidth > 0)
            {
                int overhang = Width - fixedWidth;

                if(overhang > 0)
                {
                    FFTArrayList.RemoveRange(0, overhang);
                }

                while(FFTArrayList.Count < fixedWidth)
                {
                    FFTArrayList.Insert(0, new double[Height]);
                }
            }
        }

        #endregion
    }
}