﻿using System;
using System.Linq;

namespace TestLibrary
{
    /// <summary>
    /// 도구
    /// </summary>
    public static class Tool
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 평균 FFT 배열 구하기 - GetMeanFFTArray(sff, db)

        /// <summary>
        /// 평균 FFT 배열 구하기
        /// </summary>
        /// <param name="sff">스펙트로그램 파일 포맷</param>
        /// <param name="db">데시빌 여부</param>
        /// <returns>평균 FFT 배열</returns>
        public static double[] GetMeanFFTArray(SpectrogramFileFormat sff, bool db = false)
        {
            double[] meanArray = new double[sff.FFTList[0].Length];

            foreach(double[] fftArray in sff.FFTList)
            {
                for(int y = 0; y < fftArray.Length; y++)
                {
                    meanArray[y] += fftArray[y];
                }
            }

            for(int i = 0; i < meanArray.Length; i++)
            {
                meanArray[i] /= sff.FFTList.Count();
            }

            if(db)
            {
                for(int i = 0; i < meanArray.Length; i++)
                {
                    meanArray[i] = 20 * Math.Log10(meanArray[i]);
                }
            }

            if(meanArray[meanArray.Length - 1] <= 0)
            {
                meanArray[meanArray.Length - 1] = meanArray[meanArray.Length - 2];
            }

            return meanArray;
        }

        #endregion
        #region 평균 전력 배열 구하기 - GetMeanPower(sff, db)

        /// <summary>
        /// 평균 전력 배열 구하기
        /// </summary>
        /// <param name="sff">스펙트로그램 파일 포맷</param>
        /// <param name="db">데시빌 여부</param>
        /// <returns>평균 전력 배열</returns>
        public static double[] GetMeanPower(SpectrogramFileFormat sff, bool db = false)
        {
            double[] powerArray = new double[sff.FFTList.Count];

            for(int i = 0; i < sff.FFTList.Count; i++)
            {
                powerArray[i] = (double)sff.FFTList[i].Sum() / sff.FFTList[i].Length;
            }

            if(db)
            {
                for(int i = 0; i < powerArray.Length; i++)
                {
                    powerArray[i] = 20 * Math.Log10(powerArray[i]);
                }
            }

            return powerArray;
        }

        #endregion
        #region 정점 주파수 구하기 - GetPeakFrequency(sff, firstFFTOnly)

        /// <summary>
        /// 정점 주파수 구하기
        /// </summary>
        /// <param name="sff">스펙트로그램 파일 포맷</param>
        /// <param name="firstFFTOnly">첫번째 FFT만 여부</param>
        /// <returns>정점 주파수</returns>
        public static double GetPeakFrequency(SpectrogramFileFormat sff, bool firstFFTOnly = false)
        {
            double[] frequencyArray = firstFFTOnly ? sff.FFTList[0] : GetMeanFFTArray(sff, false);

            int    peakIndex = 0;
            double peakPower = 0;

            for(int i = 0; i < frequencyArray.Length; i++)
            {
                if(frequencyArray[i] > peakPower)
                {
                    peakPower = frequencyArray[i];
                    peakIndex = i;
                }
            }

            double maximumFrequency = sff.SampleRate / 2;
            double fraction         = peakIndex / (double)sff.ImageHeight;

            if(sff.MELBINCount > 0)
            {
                double maximumMEL = FftSharp.Transform.MelFromFreq(maximumFrequency);

                return FftSharp.Transform.MelToFreq(fraction * maximumMEL);
            }
            else
            {
                return fraction * maximumFrequency;
            }
        }

        #endregion
        #region 피아노 키 구하기 - GetPianoKey(frequency)

        /// <summary>
        /// 피아노 키 구하기
        /// </summary>
        /// <param name="frequency">주파수 (단위 : HZ)</param>
        /// <returns>피아노 키</returns>
        public static int GetPianoKey(double frequency)
        {
            double pianoKey = (39.86 * Math.Log10(frequency / 440)) + 49;

            return (int)Math.Round(pianoKey);
        }

        #endregion
        #region 미디 노트 구하기 - GetMIDINote(frequency)

        /// <summary>
        /// 미디 노트 구하기
        /// </summary>
        /// <param name="frequency">주파수 (단위 : HZ)</param>
        /// <returns>미디 노트</returns>
        public static int GetMIDINote(double frequency)
        {
            return GetPianoKey(frequency) + 20;
        }

        #endregion
    }
}