﻿using System;
using System.Collections.Generic;

using NAudio.Wave;

namespace TestProject
{
    /// <summary>
    /// 리스너
    /// </summary>
    public class Listener : IDisposable
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region Field

        /// <summary>
        /// 샘플 비율
        /// </summary>
        public readonly int SampleRate;

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 웨이브 입력 이벤트
        /// </summary>
        private readonly WaveInEvent waveInEvent;

        /// <summary>
        /// 오디오 리스트
        /// </summary>
        private readonly List<double> audioList = new List<double>();

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 진폭 분수 - AmplitudeFraction

        /// <summary>
        /// 진폭 분수
        /// </summary>
        public double AmplitudeFraction { get; private set; }

        #endregion
        #region 전체 샘플 카운트 - TotalSampleCount

        /// <summary>
        /// 전체 샘플 카운트
        /// </summary>
        public double TotalSampleCount { get; private set; }

        #endregion
        #region 전체 시간 (단위 : 초) - TotalTime

        /// <summary>
        /// 전체 시간 (단위 : 초)
        /// </summary>
        public double TotalTime
        {
            get
            {
                return (double)TotalSampleCount / SampleRate;
            }
        }

        #endregion
        #region 메모리 내 샘플 카운트 - SampleCountInMemory

        /// <summary>
        /// 메모리 내 샘플 카운트
        /// </summary>
        public int SampleCountInMemory
        {
            get
            {
                return this.audioList.Count;
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - Listener(deviceIndex, sampleRate)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="deviceIndex">장치 인덱스</param>
        /// <param name="sampleRate">샘플 비율</param>
        public Listener(int deviceIndex, int sampleRate)
        {
            SampleRate = sampleRate;

            this.waveInEvent = new WaveInEvent
            {
                DeviceNumber       = deviceIndex,
                WaveFormat         = new WaveFormat(sampleRate, bits : 16, channels : 1),
                BufferMilliseconds = 20
            };

            this.waveInEvent.DataAvailable += waveInEvent_DataAvailable;

            this.waveInEvent.StartRecording();
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 신규 오디오 배열 구하기 - GetNewAudioArray()

        /// <summary>
        /// 신규 오디오 배열 구하기
        /// </summary>
        /// <returns>신규 오디오 배열</returns>
        public double[] GetNewAudioArray()
        {
            lock(this.audioList)
            {
                double[] valueArray = new double[this.audioList.Count];

                for(int i = 0; i < valueArray.Length; i++)
                {
                    valueArray[i] = this.audioList[i];
                }

                this.audioList.RemoveRange(0, valueArray.Length);

                return valueArray;
            }
        }

        #endregion
        #region 리소스 해제하기 - Dispose()

        /// <summary>
        /// 리소스 해제하기
        /// </summary>
        public void Dispose()
        {
            this.waveInEvent?.StopRecording();

            this.waveInEvent?.Dispose();
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 웨이브 입력 이벤트 데이터 이용 가능시 처리하기 - waveInEvent_DataAvailable(sender, e)

        /// <summary>
        /// 웨이브 입력 이벤트 데이터 이용 가능시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void waveInEvent_DataAvailable(object sender, WaveInEventArgs e)
        {
            int      byteCountPerSample = this.waveInEvent.WaveFormat.BitsPerSample / 8;
            int      newSampleCount     = e.BytesRecorded / byteCountPerSample;
            double[] bufferArray        = new double[newSampleCount];
            double   peak               = 0d;

            for(int i = 0; i < newSampleCount; i++)
            {
                bufferArray[i] = BitConverter.ToInt16(e.Buffer, i * byteCountPerSample);

                peak = Math.Max(peak, bufferArray[i]);
            }

            lock(this.audioList)
            {
                this.audioList.AddRange(bufferArray);
            }

            AmplitudeFraction = peak / (1 << 15);

            TotalSampleCount += newSampleCount;
        }

        #endregion
    }
}