﻿using System;
using System.Diagnostics;
using System.Drawing;
using System.Windows.Forms;

using NAudio.Wave;

using TestLibrary;

namespace TestProject
{
    /// <summary>
    /// 메인 폼
    /// </summary>
    public partial class MainForm : Form
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 스펙트로그램 제너레이터
        /// </summary>
        private SpectrogramGenerator generator;

        /// <summary>
        /// 리스너
        /// </summary>
        private Listener listener;

        /// <summary>
        /// 색상 맵 배열
        /// </summary>
        private ColorMap[] colorMapArray;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - MainForm()

        /// <summary>
        /// 생성자
        /// </summary>
        public MainForm()
        {
            InitializeComponent();

            if(WaveIn.DeviceCount == 0)
            {
                MessageBox.Show
                (
                    "No audio input devices found.\n\nThis program will now exit.",
                    "ERROR",
                    MessageBoxButtons.OK,
                    MessageBoxIcon.Error
                );

                Close();
            }

            this.deviceComboBox.Items.Clear();

            for(int i = 0; i < WaveIn.DeviceCount; i++)
            {
                this.deviceComboBox.Items.Add(WaveIn.GetCapabilities(i).ProductName);
            }

            for(int i = 9; i < 16; i++)
            {
                this.fftSizeComboBox.Items.Add($"2^{i} ({1 << i:N0})");
            }

            this.colorMapArray = ColorMap.GetColorMapArray();

            foreach(ColorMap colorMap in this.colorMapArray)
            {
                this.colorMapComboBox.Items.Add(colorMap.Name);
            }

            this.deviceComboBox.SelectedIndexChanged   += deviceComboBox_SelectedIndexChanged;
            this.fftSizeComboBox.SelectedIndexChanged  += fftSizeComboBox_SelectedIndexChanged;
            this.colorMapComboBox.SelectedIndexChanged += colorMapComboBox_SelectedIndexChanged;
            this.rollCheckBox.CheckedChanged           += rollCheckBox_CheckedChanged;
            this.resetRollButton.Click                 += resetRollButton_Click;
            this.timer.Tick                            += timer_Tick;

            this.deviceComboBox.SelectedIndex   = 0;
            this.fftSizeComboBox.SelectedIndex  = 1;
            this.colorMapComboBox.SelectedIndex = this.colorMapComboBox.Items.IndexOf("LoporaColorMap");
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Private
        //////////////////////////////////////////////////////////////////////////////// Event

        #region 장치 콤보 박스 선택 인덱스 변경시 처리하기 - deviceComboBox_SelectedIndexChanged(sender, e)

        /// <summary>
        /// 장치 콤보 박스 선택 인덱스 변경시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void deviceComboBox_SelectedIndexChanged(object sender, EventArgs e) => StartListening();

        #endregion
        #region FFT 크기 콤보 박스 선택 인덱스 변경시 처리하기 - fftSizeComboBox_SelectedIndexChanged(sender, e)

        /// <summary>
        /// FFT 크기 콤보 박스 선택 인덱스 변경시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void fftSizeComboBox_SelectedIndexChanged(object sender, EventArgs e) => StartListening();

        #endregion
        #region 색상 맵 콤보 박스 선택 인덱스 변경시 처리하기 - colorMapComboBox_SelectedIndexChanged(sender, e)

        /// <summary>
        /// 색상 맵 콤보 박스 선택 인덱스 변경시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void colorMapComboBox_SelectedIndexChanged(object sender, EventArgs e)
        {
            this.generator.Colormap = this.colorMapArray[this.colorMapComboBox.SelectedIndex];
        }

        #endregion
        #region Roll 체크 박스 체크 변경시 처리하기 - rollCheckBox_CheckedChanged(sender, e)

        /// <summary>
        /// Roll 체크 박스 체크 변경시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void rollCheckBox_CheckedChanged(object sender, EventArgs e)
        {
            this.resetRollButton.Enabled = this.rollCheckBox.Checked;
        }

        #endregion
        #region Roll 리셋하기 버튼 클릭시 처리하기 - resetRollButton_Click(sender, e)

        /// <summary>
        /// Roll 리셋하기 버튼 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void resetRollButton_Click(object sender, EventArgs e)
        {
            this.generator.ResetRoll();
        }

        #endregion
        #region 타이머 틱 처리하기 - timer_Tick(sender, e)

        /// <summary>
        /// 타이머 틱 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void timer_Tick(object sender, EventArgs e)
        {
            double[] newAudioArray = this.listener.GetNewAudioArray();

            this.generator.Add(newAudioArray, process : false);

            double multiplier = this.brightnessTrackBar.Value / 20d;

            if(this.generator.FFTCountToProcess > 0)
            {
                Stopwatch stopWatch = Stopwatch.StartNew();

                this.generator.Process();

                this.generator.SetFixedWidth(spectorgramPictureBox.Width);

                Bitmap spectrogramBitmap = new Bitmap(this.generator.Width, this.generator.Height, System.Drawing.Imaging.PixelFormat.Format32bppPArgb);

                using(Bitmap indexedSpectrogramBitmap = this.generator.GetBitmap(multiplier, this.dbCheckBox.Checked, roll : this.rollCheckBox.Checked))
                using(Graphics graphics = Graphics.FromImage(spectrogramBitmap))
                using(Pen pen = new Pen(Color.White))
                {
                    graphics.DrawImage(indexedSpectrogramBitmap, 0, 0);

                    if(this.rollCheckBox.Checked)
                    {
                        graphics.DrawLine(pen, this.generator.NextColumnIndex, 0, this.generator.NextColumnIndex, this.spectorgramPictureBox.Height);
                    }
                }

                stopWatch.Stop();

                this.spectorgramPictureBox.Image?.Dispose();

                this.spectorgramPictureBox.Image = spectrogramBitmap;

                this.renderTimeStatusLabel.Text = $"렌러딩 시간 : {stopWatch.ElapsedMilliseconds:D2} 밀리초";

                this.peakFrequencyStatusLabel.Text = $"피크 (Hz): {this.generator.GetPeak().frequencyHZ:N0}";
            }

            this.timeStatusLabel.Text = $"시간 : {this.listener.TotalTime:N3} 초";

            this.fftCountProcessedStatusLabel.Text = $"처리 FFT 수 : {this.generator.FFTCountProcessed:N0}";

            this.amplitudeProgressBar.Value = (int)(this.listener.AmplitudeFraction * this.amplitudeProgressBar.Maximum);
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Function

        #region 청취 시작하기 - StartListening()

        /// <summary>
        /// 청취 시작하기
        /// </summary>
        private void StartListening()
        {
            int sampleRate = 6000;
            int fftSize    = 1 << (9 + this.fftSizeComboBox.SelectedIndex);
            int stepSize   = fftSize / 20;

            this.spectorgramPictureBox.Image?.Dispose();

            this.spectorgramPictureBox.Image = null;

            this.listener?.Dispose();

            this.listener = new Listener(this.deviceComboBox.SelectedIndex, sampleRate);

            this.generator = new SpectrogramGenerator(sampleRate, fftSize, stepSize);

            this.spectorgramPictureBox.Height = this.generator.Height;

            this.verticalScalePictureBox.Image?.Dispose();

            this.verticalScalePictureBox.Image  = this.generator.GetVerticalScaleBitmap(this.verticalScalePictureBox.Width);
            this.verticalScalePictureBox.Height = this.generator.Height;
        }

        #endregion
    }
}